use crate::read;
use itertools::Itertools;
use std::collections::*;
use std::hash::{Hash, Hasher};
use std::io::Write;

#[derive(Debug, Clone, Eq)]
pub struct Coord {
    pub x: i32,
    pub y: usize,
    pub is_rock: bool,
}

impl Hash for Coord {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.x.hash(state);
        self.y.hash(state);
    }
}

impl PartialEq for Coord {
    fn eq(&self, other: &Self) -> bool {
        self.x == other.x && self.y == other.y
    }
}

impl Coord {
    pub fn down_left(&self) -> Self {
        Coord {
            x: self.x - 1,
            y: self.y + 1,
            is_rock: self.is_rock,
        }
    }
    pub fn down_right(&self) -> Self {
        Coord {
            x: self.x + 1,
            y: self.y + 1,
            is_rock: self.is_rock,
        }
    }
    pub fn down(&self) -> Self {
        Coord {
            x: self.x,
            y: self.y + 1,
            is_rock: self.is_rock,
        }
    }
}

pub fn run() {
    let input = read("14");
    let lines = input.lines();
    let mut rocks: HashSet<Coord> = lines
        .flat_map(|line| {
            line.split(" -> ")
                .map(|c| {
                    let nums = c.split_once(',').unwrap();
                    Coord {
                        x: nums.0.parse().unwrap(),
                        y: nums.1.parse().unwrap(),
                        is_rock: true,
                    }
                })
                .tuple_windows()
                .flat_map(|(s, e)| {
                    let mut path = vec![];
                    for nx in s.x..=e.x {
                        for ny in s.y..=e.y {
                            path.push(Coord {
                                x: nx,
                                y: ny,
                                is_rock: true,
                            })
                        }
                    }
                    for nx in e.x..=s.x {
                        for ny in e.y..=s.y {
                            path.push(Coord {
                                x: nx,
                                y: ny,
                                is_rock: true,
                            })
                        }
                    }
                    path
                })
                .collect_vec()
        })
        .collect();

    #[cfg(feature = "part1")]
    {
        println!("Day 14, Part 01: {:?}", p1(&mut rocks.clone()));
    }

    #[cfg(feature = "part2")]
    {
        let floor = rocks.iter().map(|r| r.y).max().unwrap() + 2;
        for x in -1000..=1000 {
            rocks.insert(Coord {
                x,
                y: floor,
                is_rock: true,
            });
        }
        println!("Day 14, Part 02: {:?}", p2(&mut rocks.clone()));
    }
}

fn p1(rocks: &mut HashSet<Coord>) -> usize {
    let bottom = rocks.iter().map(|r| r.y).max().unwrap();
    let sand_start = Coord {
        x: 500,
        y: 0,
        is_rock: false,
    };
    let mut count_p1 = 0;

    loop {
        let mut sand = sand_start.clone();

        loop {
            //print(rocks, &sand);
            if sand.y > bottom {
                return count_p1;
            }
            let d = sand.down();
            let dl = sand.down_left();
            let dr = sand.down_right();

            if rocks.contains(&d) {
                if rocks.contains(&dl) {
                    if rocks.contains(&dr) {
                        count_p1 += 1;
                        rocks.insert(sand.clone());
                        break;
                    }
                    sand = dr;
                } else {
                    sand = dl;
                }
            } else {
                sand = d;
            }
        }
    }
}

fn p2(rocks: &mut HashSet<Coord>) -> usize {
    let sand_start = Coord {
        x: 500,
        y: 0,
        is_rock: false,
    };
    let mut count_p2 = 0;

    loop {
        let mut sand = sand_start.clone();
        if rocks.contains(&sand) {
            return count_p2;
        }

        loop {
            let d = sand.down();
            let dl = sand.down_left();
            let dr = sand.down_right();

            if rocks.contains(&d) {
                if rocks.contains(&dl) {
                    if rocks.contains(&dr) {
                        count_p2 += 1;
                        rocks.insert(sand.clone());
                        break;
                    }
                    sand = dr;
                } else {
                    sand = dl;
                }
            } else {
                sand = d;
            }
        }
    }
}

fn print(rocks: &HashSet<Coord>, sand: &Coord) {
    let min_y = 10;
    let min_x = 452;
    let max_y = 180;
    let max_x = 514;

    let bounds = " ".repeat((max_y as usize) - (min_y as usize));
    let mut out = format!("\n{}\n", bounds);

    for x in (min_x..=max_x).rev() {
        for y in min_y..=max_y {
            let testing = &Coord {
                x,
                y,
                is_rock: false,
            };
            let c = match (sand == testing, rocks.get(testing)) {
                (true, _) => 'o',
                (_, None) => ' ',
                (_, Some(coord)) if coord.is_rock => '#',
                _ => 'o',
            };
            out += &format!("{}", c);
        }
        out += "\n";
    }
    out += &bounds;
    print!("{}", out);
    std::io::stdout().flush().unwrap();
    std::thread::sleep(std::time::Duration::from_millis(60));
}
